/*
 * $Header: /home/cvs/jakarta-slide/src/stores/org/apache/slide/store/txfile/XMLResourceDescriptor.java,v 1.5.2.3 2004/04/01 06:55:05 ozeigermann Exp $
 * $Revision: 1.5.2.3 $
 * $Date: 2004/04/01 06:55:05 $
 *
 * ====================================================================
 *
 * Copyright 1999-2002 The Apache Software Foundation 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

package org.apache.slide.store.txfile;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import org.apache.slide.common.*;
import org.apache.slide.lock.NodeLock;
import org.apache.slide.structure.*;
import org.apache.slide.content.*;

import org.jdom.*;
import org.jdom.input.*;
import org.jdom.output.*;

import org.apache.slide.store.txfile.rm.ResourceManagerException;
import org.apache.slide.store.txfile.rm.impl.FileResourceManager;
import org.apache.slide.util.logger.*;

import slidestore.file.UriProperties;

import java.io.File;

/**
 * XML descriptor as a resource in a {@link FileResourceManager}. Copies code from
 * UriProperties, that's why Marc is listed as author as well.
 * 
 * @author <a href="mailto:ozeigermann@c1-fse.de">Oliver Zeigermann</a>
 * @author <a href="mailto:mdecugis@mds-desing.fr">Marc Dcugis</a>
 * @see FileResourceManager
 */
public class XMLResourceDescriptor extends UriProperties {

    private static String tmpDir;
    static {
        try {
            tmpDir = File.createTempFile("xyz", null).getParent();
        } catch (IOException e) {
            tmpDir = null;
        }
    }

    protected static final String PATH_EXTENSION = ".def.xml";
    protected final FileResourceManager rm;
    protected final TxXMLFileDescriptorsStore store;
    protected final String characterEncoding;
    protected volatile Object txId;
    protected volatile String uri;
    protected final String loadPath;

    protected StoreLogger logger;

    /**
     * Creates an XML descriptor resource.
     * 
     * @param uri uri of the resource
     * @param rm resource manager to load / store this descriptor from / to
     * @param txId identifier for the transaction in which the descriptor is to be managed
     * @param characterEncoding charcter enconding used to store this descriptor in XML 
     * @throws ServiceAccessException if anything goes wrong at system level
     * @throws ObjectNotFoundException if <code>load</code> is set to <code>true</code> and the descriptor is not present in resource manager
     */
    public XMLResourceDescriptor(
        Uri uri,
        TxXMLFileDescriptorsStore store,
        FileResourceManager rm,
        Object txId,
        String characterEncoding)
        throws ServiceAccessException, ObjectNotFoundException {
        // XXX super class tries to create folders, no way to stop it:
        // at least let it create folders in a place where it can do no harm
        super(tmpDir, uri);

        logger = rm.getLogger().cloneWithNewLogChannel(XMLResourceDescriptor.class.getName());

        this.rm = rm;
        this.store = store;
        this.characterEncoding = characterEncoding;
        this.txId = txId;

        if (uri == null) {
            store.throwInternalError("Trying to initialize XMLResourceDescriptor with null URI");
        }
        this.uri = uri.toString();

        this.loadPath = uri.toString() + PATH_EXTENSION;
        load();
    }

    /**
     * Stores this descriptor to the resource manager.
     * 
     * @throws ServiceAccessException if anything goes wrong at system level
     * @throws ObjectNotFoundException if the descriptor has not been created before 
     */
    public synchronized void save() throws ServiceAccessException, ObjectNotFoundException {
        if (txId == null) {
            store.throwInternalError("Not inside tx");
        }
        logger.logFine("Tx " + txId + " saves data for " + loadPath);

        OutputStream os = null;
        try {
            os = rm.writeResource(txId, loadPath);
            save(os);
        } catch (ResourceManagerException e) {
            if (e.getStatus() == ResourceManagerException.ERR_NO_SUCH_RESOURCE) {
                throw new ObjectNotFoundException(uri);
            } else {
                store.throwInternalError(e, uri);
            }
        } finally {
            try {
                if (os != null) {
                    os.close();
                }
            } catch (IOException e) {
            }
        }
    }

    /**
     * Creates this descriptor in the resource manager.
     * 
     * @throws ServiceAccessException if anything goes wrong at system level
     * @throws ObjectAlreadyExistsException if the descriptor already exists 
     */
    public synchronized void create() throws ServiceAccessException, ObjectAlreadyExistsException {
        logger.logFiner("Tx " + txId + " creates " + loadPath);
        if (txId == null) {
            store.throwInternalError("Not inside tx");
        }
        try {
            rm.createResource(txId, loadPath, false);
            init();
        } catch (ResourceManagerException e) {
            if (e.getStatus() == ResourceManagerException.ERR_RESOURCE_EXISTS) {
                throw new ObjectAlreadyExistsException(uri.toString());
            } else {
                store.throwInternalError(e, uri);
            }
        }
    }

    /**
     * Deletes this descriptor from the resource manager.
     * 
     * @throws ServiceAccessException if anything goes wrong at system level
     * @throws ObjectNotFoundException if the descriptor does not exist 
     */
    public synchronized void delete() throws ServiceAccessException, ObjectNotFoundException {
        logger.logFiner("Tx " + txId + " deletes " + loadPath);
        if (txId == null) {
            store.throwInternalError("Not inside tx");
        }
        try {
            rm.deleteResource(txId, loadPath, false);
            object = null;
        } catch (ResourceManagerException e) {
            if (e.getStatus() == ResourceManagerException.ERR_NO_SUCH_RESOURCE) {
                throw new ObjectNotFoundException(uri.toString());
            } else {
                store.throwInternalError(e, uri);
            }
        }
    }

    /**
     * Loads this descriptor from the resource manager.
     * 
     * @throws ServiceAccessException if anything goes wrong at system level
     * @throws ObjectNotFoundException if the descriptor does not exist 
     */
    public synchronized void load() throws ServiceAccessException, ObjectNotFoundException {
        logger.logFiner("Tx " + txId + " loads data for " + loadPath);

        InputStream is = null;
        try {
            logger.logFinest("Faking read access from outside tx for " + loadPath);
            if (txId != null) {
                if (rm.resourceExists(txId, loadPath)) {
                    is = rm.readResource(txId, loadPath);
                    load(is);
                } else {
                    init();
                }
            } else {
                if (rm.resourceExists(loadPath)) {
                    is = rm.readResource(loadPath);
                    load(is);
                } else {
                    init();
                }
            }
        } catch (ResourceManagerException e) {
            if (e.getStatus() == ResourceManagerException.ERR_NO_SUCH_RESOURCE) {
                throw new ObjectNotFoundException(uri);
            } else {
                store.throwInternalError(e, uri);
            }
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
            } catch (IOException e) {
            }
        }
    }

    /**
     * Gets the URI of this descriptor.
     * 
     * @return the URI
     */
    public String getUri() {
        return uri;
    }

    /**
     * Gets the transaction this descriptor lives in.
     * 
     * @return the transaction identifier
     */
    public Object getTxId() {
        return txId;
    }

    /**
     * Checks if the specified object is a descriptor with the same URI in the same transaction.
     * 
     * 
     * @param o object to compare this descriptor to
     * @return <code>true</code> if object is equal as described above 
     */
    public synchronized boolean equals(Object o) {
        return (
            this == o
                || (o != null
                    && o instanceof XMLResourceDescriptor
                    && ((XMLResourceDescriptor) o).uri.equals(uri)
                    && ((XMLResourceDescriptor) o).txId.equals(txId)));
    }

    public synchronized String toString() {
        return txId + ": " + uri;
    }

    protected synchronized void save(OutputStream os) throws ServiceAccessException {
        Element aRoot = encode();
        Document aDocument = new Document(aRoot);

        XMLOutputter aOutputter = new XMLOutputter();
        aOutputter.setEncoding(characterEncoding);
        aOutputter.setNewlines(true);
        aOutputter.setIndent("  ");
        aOutputter.setLineSeparator("\n");
        try {
            aOutputter.output(aDocument, os);
            os.flush();
        } catch (IOException e) {
            store.throwInternalError(e);
        }
    }

    public synchronized Element encodeObject() {
        if (object != null) {
            return super.encodeObject();
        } else {
            Element aElementObjectNode = new Element("objectnode");
            // for null locks
            aElementObjectNode.setAttribute("classname", "null");
            aElementObjectNode.setAttribute("uri", uri.toString());
            aElementObjectNode.addContent(new Element("children"));
            aElementObjectNode.addContent(new Element("links"));
            return aElementObjectNode;
        }
    }

    public synchronized void decodeObject(Element aElement) throws ServiceAccessException {
        Element aElementObjectNode = aElement.getChild("objectnode");
        String aClasseName = aElementObjectNode.getAttributeValue("classname");
        if (!"null".equals(aClasseName)) {
            super.decodeObject(aElement);
            uri = object.getUri();
        } else {
            object = null;
            uri = aElementObjectNode.getAttributeValue("uri");
        }
    }

    public synchronized void decodePermissions(Element aElement) {
        permissions = new Vector();
        Element aPermissions = aElement.getChild("permissions");
        List aList = aPermissions.getChildren();
        for (int i = 0; i < aList.size(); i++) {
            Element aChild = (Element) aList.get(i);
            permissions.addElement(decodePermission(aChild, uri));
        }
    }

    public synchronized void decodeLocks(Element aElement) throws ServiceAccessException {
        try {
            locks = new Vector();
            Element aElementLocks = aElement.getChild("locks");
            List aList = aElementLocks.getChildren();

            for (int i = 0; i < aList.size(); i++) {
                Element aChild = (Element) aList.get(i);
                String aSubject = aChild.getAttributeValue("subjectUri");
                String aType = aChild.getAttributeValue("typeUri");
                Date aDateExpiration = dateFormat.parse(aChild.getAttributeValue("date"));
                boolean aInheritable = new Boolean(aChild.getAttributeValue("inheritance")).booleanValue();
                boolean aNegative = new Boolean(aChild.getAttributeValue("exclusive")).booleanValue();
                String aLockId = aChild.getAttributeValue("lockId");
                String ownerInfo = aChild.getAttributeValue("owner");

                locks.addElement(new NodeLock(aLockId, uri, aSubject, aType, aDateExpiration, aInheritable, aNegative, ownerInfo));
            }
        } catch (Exception e) {
            e.printStackTrace();
            throw new ServiceAccessException(null, e);
        }

    }

    public synchronized void decodeRevisionDescriptors(Element aElement) {
        Element aRevisionsHistory = aElement.getChild("revisionsHistory");

        NodeRevisionNumber aInitialRevision =
            decodeRevisionNumber(aRevisionsHistory.getAttributeValue("initialRevision"));
        boolean aUseVersionning = new Boolean(aRevisionsHistory.getAttributeValue("useVersioning")).booleanValue();

        Element aBranchesElement = aRevisionsHistory.getChild("branches");
        if (aBranchesElement == null) {
            revisionDescriptors =
                new NodeRevisionDescriptors(
                    uri,
                    aInitialRevision,
                    new Hashtable(),
                    new Hashtable(),
                    new Hashtable(),
                    aUseVersionning);
            return;
        }

        List aList = aBranchesElement.getChildren();
        Hashtable aLastestRevisions = new Hashtable();
        for (int i = 0; i < aList.size(); i++) {
            Element aChild = (Element) aList.get(i);
            String aName = aChild.getAttributeValue("name");
            NodeRevisionNumber aRevisionNumber = decodeRevisionNumber(aChild.getAttributeValue("lastestRevision"));
            aLastestRevisions.put(aName, aRevisionNumber);
        }
        Hashtable aBranches = new Hashtable();
        Element aRevisionsElement = aRevisionsHistory.getChild("revisions");
        aList = aRevisionsElement.getChildren();
        for (int i = 0; i < aList.size(); i++) {
            Element aChild = (Element) aList.get(i);
            NodeRevisionNumber aStartNumber = decodeRevisionNumber(aChild.getAttributeValue("start"));
            List aSuccessors = aChild.getChildren();
            Vector aSuccessorsNumbers = new Vector();
            for (int k = 0; k < aSuccessors.size(); k++) {
                Element aSuccessor = (Element) aSuccessors.get(k);
                NodeRevisionNumber aRevisionNumber = decodeRevisionNumber(aSuccessor.getAttributeValue("number"));
                aSuccessorsNumbers.addElement(aRevisionNumber);
            }
            aBranches.put(aStartNumber, aSuccessorsNumbers);
        }
        revisionDescriptors =
            new NodeRevisionDescriptors(
                uri,
                aInitialRevision,
                new Hashtable(),
                aLastestRevisions,
                aBranches,
                aUseVersionning);
    }

    public synchronized void decodeRevisionDescriptor(Element aParent) {
        descriptor = new Hashtable();

        Element aElement = aParent.getChild("descriptor");
        if (aElement == null)
            return;

        List aList = aElement.getChildren();

        for (int i = 0; i < aList.size(); i++) {
            Element aChild = (Element) aList.get(i);
            String aBranchName = aChild.getAttributeValue("branchName");
            NodeRevisionNumber aRevisionNumber = decodeRevisionNumber(aChild.getAttributeValue("number"));

            Vector aLabels = new Vector();
            Element aLabelsElement = (Element) aChild.getChild("labels");
            List aLabelList = aLabelsElement.getChildren();
            for (int k = 0; k < aLabelList.size(); k++) {
                Element aLabel = (Element) aLabelList.get(k);
                aLabels.addElement(aLabel.getAttributeValue("val"));
            }

            Hashtable aProperties = new Hashtable();
            Element aPropertiesElement = (Element) aChild.getChild("properties");
            List aPropertiesList = aPropertiesElement.getChildren();
            for (int k = 0; k < aPropertiesList.size(); k++) {
                Element aProperty = (Element) aPropertiesList.get(k);
                NodeProperty aProp = decodeNodeProperty(aProperty, uri);
                String key = aProperty.getAttributeValue("namespace") + aProperty.getAttributeValue("name");
                aProperties.put(key, aProp);
            }
            NodeRevisionDescriptor aNode =
                new NodeRevisionDescriptor(aRevisionNumber, aBranchName, aLabels, aProperties);
            descriptor.put(aRevisionNumber.toString(), aNode);
        }
    }

    protected synchronized void load(InputStream is) throws ServiceAccessException {
        SAXBuilder aBuilder = new SAXBuilder();
        try {
            Document aDocument = aBuilder.build(is);
            decode(aDocument.getRootElement());
        } catch (JDOMException e) {
            store.throwInternalError(e);
        } catch (IOException e) {
            store.throwInternalError(e);
        }
    }

    protected synchronized void init() throws ServiceAccessException {
        // need to set this null, as AbstractUriProperties.retrieveObject relies on it
        object = null;
        permissions = new Vector();
        locks = new Vector();
        revisionDescriptors = new NodeRevisionDescriptors();
        descriptor = new Hashtable();
    }

    protected synchronized void save(Uri uri) throws ServiceAccessException {
        if (txId == null) {
            store.throwInternalError("Not inside tx");
        }
        // null objects must be saveable, as it is possible to have nulllocks 

        //        if (object == null) {
        //            store.throwInternalError("Resource '" + uri + "' not initialized");
        //        }
        try {
            save();
        } catch (ObjectNotFoundException e) {
            store.throwInternalError(e);
        }
    }

}